use tokio::sync::mpsc::UnboundedSender;
use color_eyre::Result;
use subxt::{
    backend::{
        legacy::LegacyRpcMethods, 
        rpc::RpcClient,
    }, 
    tx::{TxProgress, TxStatus}, 
    utils::H256, 
    OnlineClient,
};

mod legacy_rpc_calls;
mod predefinded_calls;
mod subscriptions;

use crate::{
    types::ActionLevel,
    action::Action, 
    casper::CasperConfig,
};

pub use subscriptions::{FinalizedSubscription, BestSubscription};

pub struct Network {
    action_tx: UnboundedSender<Action>,
    online_client_api: OnlineClient<CasperConfig>,
    legacy_client_api: LegacyRpcMethods<CasperConfig>,
    rpc_client: RpcClient,
    best_hash: Option<H256>,
    finalized_hash: Option<H256>,
    accounts_to_watch: std::collections::HashSet<[u8; 32]>,
    transactions_to_watch: Vec<TxProgress<CasperConfig, OnlineClient<CasperConfig>>>,
}

impl Network {
    pub fn new(
        action_tx: UnboundedSender<Action>, 
        online_client_api: OnlineClient<CasperConfig>,
        legacy_client_api: LegacyRpcMethods<CasperConfig>,
        rpc_client: RpcClient,
    ) -> Self {
        Self { 
            action_tx,
            online_client_api,
            legacy_client_api,
            rpc_client,
            best_hash: None,
            finalized_hash: None,
            accounts_to_watch: Default::default(),
            transactions_to_watch: Default::default(),
        }
    }

    pub async fn handle_network_event(&mut self, io_event: Action) -> Result<()> {
        match io_event {
            Action::NewBestHash(hash) => {
                self.best_hash = Some(hash);
                for account_id in self.accounts_to_watch.iter() {
                    predefinded_calls::get_balance(&self.action_tx, &self.online_client_api, &account_id).await?;
                }
                Ok(())
            },
            Action::NewFinalizedHash(hash) => {
                self.finalized_hash = Some(hash);
                let length = self.transactions_to_watch.len();
                for i in (0..length).rev() {
                    let pending_tx = &mut self.transactions_to_watch[i];
                    let ext_hash = pending_tx.extrinsic_hash();
                    match (*pending_tx).next().await {
                        Some(Ok(status)) => {
                            match status {
                                TxStatus::Validated => self.action_tx.send(Action::WalletLog(format!("transaction {} is part of future queue", ext_hash), ActionLevel::Info))?,
                                TxStatus::Broadcasted { num_peers } => self.action_tx.send(Action::WalletLog(format!("transaction {} has been broardcasted to {} nodes", ext_hash, num_peers), ActionLevel::Info))?,
                                TxStatus::NoLongerInBestBlock => self.action_tx.send(Action::WalletLog(format!("transaction {} is no longer in a best block", ext_hash), ActionLevel::Warn))?,
                                TxStatus::InBestBlock(b) => self.action_tx.send(Action::WalletLog(format!("transaction {} included in the block header {}", b.extrinsic_hash(), b.block_hash()), ActionLevel::Info))?,
                                TxStatus::InFinalizedBlock(b) => {
                                    self.action_tx.send(Action::WalletLog(format!("transaction {} has been finalized in block header {}", b.extrinsic_hash(), b.block_hash()), ActionLevel::Info))?;
                                    self.transactions_to_watch.remove(i);
                                }
                                TxStatus::Error { message } => {
                                    self.action_tx.send(Action::WalletLog(format!("transaction {} error, something get wrong: {message}", ext_hash), ActionLevel::Error))?;
                                    self.transactions_to_watch.remove(i);
                                }
                                TxStatus::Invalid { message } => {
                                    self.action_tx.send(Action::WalletLog(format!("transaction {} invalid: {message}", ext_hash), ActionLevel::Error))?;
                                    self.transactions_to_watch.remove(i);
                                }
                                TxStatus::Dropped { message } => {
                                    self.action_tx.send(Action::WalletLog(format!("transaction {} was dropped: {message}", ext_hash), ActionLevel::Error))?;
                                    self.transactions_to_watch.remove(i);
                                }
                            }
                        },
                        _ => {
                            self.action_tx.send(Action::WalletLog(format!("transaction {} was dropped", ext_hash), ActionLevel::Error))?;
                            self.transactions_to_watch.remove(i);
                        }
                    }
                }
                Ok(())
            },
            Action::GetSystemHealth => legacy_rpc_calls::get_system_health(&self.action_tx, &self.legacy_client_api).await,
            Action::GetNodeName => legacy_rpc_calls::get_node_name(&self.action_tx, &self.legacy_client_api).await,
            Action::GetGenesisHash => legacy_rpc_calls::get_genesis_hash(&self.action_tx, &self.legacy_client_api).await,
            Action::GetChainName => legacy_rpc_calls::get_chain_name(&self.action_tx, &self.legacy_client_api).await,
            Action::GetChainVersion => legacy_rpc_calls::get_system_version(&self.action_tx, &self.legacy_client_api).await,
            Action::GetBlockAuthor(hash, logs) => predefinded_calls::get_block_author(&self.action_tx, &self.online_client_api, &logs, &hash).await,
            Action::GetActiveEra => predefinded_calls::get_active_era(&self.action_tx, &self.online_client_api).await,
            Action::GetEpochProgress => predefinded_calls::get_epoch_progress(&self.action_tx, &self.online_client_api).await,
            Action::GetPendingExtrinsics => predefinded_calls::get_pending_extrinsics(&self.action_tx, &self.rpc_client).await,

            Action::GetExistentialDeposit => predefinded_calls::get_existential_deposit(&self.action_tx, &self.online_client_api).await,
            Action::GetTotalIssuance => predefinded_calls::get_total_issuance(&self.action_tx, &self.online_client_api).await,

            Action::BalanceRequest(account_id, remove) => {
                if remove {
                    let _ = self.accounts_to_watch.remove(&account_id);
                    Ok(())
                } else {
                    let _ = self.accounts_to_watch.insert(account_id);
                    predefinded_calls::get_balance(&self.action_tx, &self.online_client_api, &account_id).await
                }
            }
            Action::TransferBalance(sender, receiver, amount) => {
                let sender: [u8; 32] = hex::decode(sender)
                    .expect("stored seed is valid hex string; qed")
                    .as_slice()
                    .try_into()
                    .expect("stored seed is valid length; qed");

                if let Ok(tx_progress) = predefinded_calls::transfer_balance(
                    &self.action_tx, 
                    &self.online_client_api, 
                    &sender, 
                    &receiver, 
                    &amount,
                ).await {
                    self.transactions_to_watch.push(tx_progress);
                }
                Ok(())
            }
            _ => Ok(())
        }
    }
}
